#include "main.h"

int fd_events_log;
int fd_pipes_log;
int N;
pipe_fd * pipes;

int main(int argc, char * argv[])
{
  int X = check_args(argc, argv);

  if (X < 1 || X > 9)
    {
      log_error(0);
    }
  
  N = X + 1;

  fd_events_log = open(evengs_log, O_WRONLY | O_APPEND | O_CREAT | O_TRUNC, 0777);
  if (fd_events_log < 0)
    log_error(1);
  fd_pipes_log = open(pipes_log, O_WRONLY | O_APPEND | O_CREAT | O_TRUNC, 0777);
  if (fd_pipes_log < 0)
    log_error(1);

  // open pipes
  open_pipes();
  pid_t all_pids[N];

  all_pids[0] = getpid();
  // fork processes
  for (local_id i = 1; i < N; i++)
    {
      all_pids[i] = fork();
      switch(all_pids[i])
	{
	case 0:// child
	  pipe_work(i);
	  exit(0);
	  break;
	case -1: // error
	  log_error(2);
	  break;
	default: // parent
	  break;
	}
    }
  
  parent_work();
  wait_children(all_pids);

  close(fd_events_log);
  close(fd_pipes_log);

  exit(0);
 }

void log_error(int err_num)
{
  switch(err_num)
    {
    case 0: // show ussage
      printf("%s", help_ussage_msg);
      exit(1);
      break;
    case 1:
      printf("%s", open_file_log_error_msg);
      exit(1);
      break;
    case 2:
      printf("%s", fork_error);
      exit(1);
      break;
    }
}

int check_args(int argc, char * argv[])
{
  if (argc < 3 || strcmp(argv[1], p_key) != 0)
    {
      return 0;
    }  

  return atoi(argv[2]);
}

void open_pipes()
{
  pipes = (pipe_fd *)malloc(N * (N - 1) * sizeof(pipe_fd));
  pipe_fd * tmp;

  int fds[2];

  for (int i = 0; i < N; i++)
    for (int j = 0; j < N; j++)
      {
	if (i != j)
	  {
	    pipe(fds); 
	    tmp = get_struct(i, j);
	    tmp->fd_write = fds[1];
	    log_pipe(0, i, j);

	    tmp = get_struct(j, i);
	    tmp->fd_read = fds[0];
	    log_pipe(0, j, i);
	  }
      }
}

void close_redundant_pipes(local_id id)
{
  pipe_fd * tmp;

  for (local_id i = 0; i < N; i++)
  {
    if (i == id) continue;

    for (local_id j = 0; j < N; j++)
      {
	if (i != j)
	  {
	    tmp = get_struct(i, j);
	    
	    close(tmp->fd_write);	 
	    close(tmp->fd_read);
	    log_pipe(1, i, j);
	  }
      }
  }
}

void close_self_pipes(local_id id)
{
  pipe_fd * tmp;

  for (local_id i = 0; i < N; i++)
    {
      if (i != id)
	{
	  tmp = get_struct(id, i);
	  close(tmp->fd_write);
	  close(tmp->fd_read);
	  log_pipe(1, id, i);
	}
    }
}

// formirate second index in memory (shift)
// by indexes in matrix 
int f_index(int i, int j)
{
  if ( j >= i)
    return j - 1;
  else
    return j;
}

pipe_fd* get_struct(int from, int to)
{
  int index = from * (N - 1) + f_index(from, to);
  pipe_fd *result = pipes + (size_t)index;
  return result;
}

void parent_work()
{
  local_id id = 0;
  close_redundant_pipes(id);
  receive_any(&id, NULL);
  wait_children();
  close_self_pipes(id);
  free(pipes);
}

void pipe_work(local_id id)
{
  close_redundant_pipes(id);
  synchronize_other_start(id);
  computation(id);
  synchronize_other_done(id);
  close_self_pipes(id);
}

void synchronize_other_start(local_id id)
{
  log_event(0, id);
  
  Message m;

  int len = sprintf(m.s_payload, log_started_fmt, id, getpid(), getppid());

  m.s_header.s_magic = MESSAGE_MAGIC;
  m.s_header.s_payload_len = len;
  m.s_header.s_type = STARTED;
  m.s_header.s_local_time = time(NULL);
   
  // send message
  if (send_multicast(&id, &m) != 0)
    {      
      exit(1);
    }
 
  // wait other message
  receive_any(&id, NULL);
  log_event(1, id);
}

void computation(local_id id)
{

}

void synchronize_other_done(local_id id)
{
  log_event(2, id);

  Message m;

  int len = sprintf(m.s_payload, log_done_fmt, id);

  m.s_header.s_magic = MESSAGE_MAGIC;
  m.s_header.s_payload_len = len;
  m.s_header.s_type = DONE;
  m.s_header.s_local_time = time(NULL);

  // send message
  if (send_multicast(&id, &m) != 0)
    {
      exit(1);
    }

  // receive other messages
  receive_any(&id, NULL);
  log_event(3, id);
}

int send_multicast(void * self, const Message * msg)
{
  local_id l_id = *(local_id *)self;

  for (local_id i = 0; i < N; i++)
    {
      if (i != l_id) 
	if (send(self, i, msg) != 0)
	  {
	    return 1; 
	  }
    }
  return 0;
}

int send(void * self, local_id dst, const Message * msg)
{
  local_id l_id = *(local_id *)self;

  pipe_fd * tmp = get_struct(l_id, dst);

  int result = write(tmp->fd_write, msg, sizeof(MessageHeader) + msg->s_header.s_payload_len);
  
  if (result > 0)
    return 0;
  else
    {
      return 1;
    }
}

int receive_any(void * self, Message * msg)
{
  local_id l_id = *(local_id *)self;
  Message m;

  for (local_id i = 1; i < N; i++)
      if (i != l_id)
	if (receive(&l_id, i, &m) != 0)
	  return 1;

  return 0;
}

int receive (void * self, local_id from, Message * msg)
{
  local_id id = *(local_id *)self;
  pipe_fd *tmp = get_struct(id, from);

  MessageHeader mh;
  if (read(tmp->fd_read, &mh, sizeof(MessageHeader)) != 0)
      return 1;

  int len = mh.s_payload_len;
  char buf[len];
  if (read(tmp->fd_read, buf, len) != 0)
    return 1;

  msg->s_header = mh;
  strcpy(msg->s_payload, buf);

  return 0;
}

void log_event(int num_event, local_id l_id)
{
  char buf[100];

  switch(num_event)
    {
    case 0:
      sprintf(buf, log_started_fmt, l_id, getpid(), getppid());
      break;
    case 1:
      sprintf(buf, log_received_all_started_fmt, l_id);
      break;
    case 2:
      sprintf(buf, log_done_fmt, l_id);
      break;
    case 3:
      sprintf(buf, log_received_all_done_fmt, l_id);
    }

  printf(buf, 0);
  write(fd_events_log, buf, strlen(buf));
}

void log_pipe(int num_event, int from, int to)
{
  char buf[100];

  switch (num_event)
    {
    case 0:
      sprintf(buf, pipe_opend_msg, from, to);
      write(fd_pipes_log, buf, strlen(buf));
      break;
    case 1:
      sprintf(buf, pipe_closed_msg, from, to);
      write(fd_pipes_log, buf, strlen(buf));
      break;
    }
}

void wait_children()
{
  while(wait(NULL) > 0)
    {
      
    }
}
